#define __NEW_STARLET 1
#include "version.h"
#pragma module WATCH_DIR SW_VERSION

/*
**++
**
** Copyright  1994 James F. Duff.  ALL RIGHTS RESERVED.
**
** License Terms and Conditions:
**
** You may freely copy and distribute verbatim copies of this software package
** as you receive it, in any medium, provided that you conspicuously and
** appropriately publish on each copy this copyright notice and
** disclaimer of warranty; keep intact all the notices that refer to these
** license terms and conditions and to the absence of any warranty; and give
** any other recipients of the software package a copy of these license 
** terms and conditions along with the software.
**
** You may modify the software for your own use, but you MAY NOT distribute
** modified copies of the software.
**
** By copying or distributing the software package you indicate your
** acceptance of this license to do so, and all its terms and conditions.
**
** Each time you redistribute the software package the recipient
** automatically receives a license from the original licensor to copy,
** distribute or modify the software package subject to these terms and
** conditions. You may not impose any further restrictions on the recipients'
** exercise of the rights granted herein.
**
** BECAUSE THE SOFTWARE IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
** FOR THE SOFTWARE, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT WHEN
** OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDER PROVIDE THE SOFTWARE
** "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,
** INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
** AND FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY
** AND PERFORMANCE OF THE SOFTWARE IS WITH YOU.  SHOULD THE SOFTWARE PROVE
** DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
** CORRECTION.
**
** YOU WARRANT THAT IN NO EVENT WILL THE COPYRIGHT HOLDER BE LIABLE TO
** YOU FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR
** CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE
** SOFTWARE (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED
** INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF THE
** SOFTWARE TO OPERATE WITH ANY OTHER SOFTWARE, FIRMWARE OR HARDWARE), EVEN IF
** SUCH HOLDER HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
**
** NO TITLE TO AND OWNERSHIP OF THE SOFTWARE PACKAGE IS HEREBY TRANSFERRED.
**
** Send suggestions, bug reports, money ;-) to jim@eight-cubed.com
**
** Now that's out of the way, on with the code...
**
**--
*/

/*
**++
**
**  MODULE DESCRIPTION:
**
**	Did you ever write a program that would periodically poll a directory
**	to see if there are files to process and wonder if there was a better
**	way?  Well, this is it!
**
**	This program demonstrates how to take out what is known as the
**	"Serialization lock" for a file.  This lock is described in the
**	"VMS File System Interals" manual.  This lock is taken out before
**	the file system attempts to write to the directory file, and, as
**	this program has already taken it out in protected write mode, a
**	blocking AST fires which drops the lock (allowing the file system
**	to get it) and awakens the main module.
**
**	How to use this code:
**
**	Set the executable up as a foreign command.  For example, do the
**	following (assuming the executable is in a directory pointed to by
**	logical EXE:):
**
**		$ watch_dir :== $exe:watch_dir.exe
**
**	You must specify the name of the directory to watch as the sole
**	parameter.  If you run the program instead of setting it up as a
**	foreign command, the program will prompt you for the directory spec.
**	You can specify the directory either by specifying the filename of the
**	directory, or you can use normal directory naming syntax.  For example,
**	if you want to watch for changes in contents of a directory called
**	TEST.DIR that is in the MFD of DSA1:, you could use either of the
**	following commands:
**
**		$ watch_dir DSA1:[000000]TEST.DIR
**		$ watch_dir DSA1:[000000.TEST]
**
**	If you specify the filename method (ie, DSA1:[000000]TEST.DIR) the
**	utility makes sure that the filename you specify is really a directory.
**	For the other case, RMS can confirm that the directory exists.
**
**	When something accesses the directory, the utility will exit to DCL.
**
**	You cannot use this code to watch for specific files in a directory
**	(although it would be a simple matter to add code to the main routine
**	to search for the file you were interested in, and go back to sleep
**	if the main code was awakened and the file of interest was not
**	present).
**
**	Please, please, please understand that this code calls routines in
**	kernel mode that (potentially) interfer with what the file system is
**	doing and could potentially crash the operating system if you do
**	something strange (like modifying this code, or somehow passing in
**	strange parameters).  There should be sufficient blade guards on this
**	code so that's difficult, but anything is possible if you have
**	the privilege to run this code.
**
**	Do not run the code under control of the the user-mode debugger.  Do
**	understand what is going on in here _before_ you run the code.  If
**	you don't understand, and have questions, please drop me a line
**	_before_ you run the code.  Please re-read the disclaimer before
**	running the code.  This module is certainly not one to experiment
**	with on a production box.  You have been warned.
**
**  AUTHORS:
**
**      James F. Duff
**
**  CREATION DATE:
**
**	05-Sep-1994
**
**
**  ENVIRONMENT:
**
**	User and Kernel modes.  SYSLCK and CMKRNL privileges required.
**
**
**  MODIFICATION HISTORY:
**
**      X01-00	    Jim Duff	    05-Sep-1994
**	Original version of module
**
**	X01-01	    Jim Duff	    26-Jan-2003
**	Do some code cleanup.
**
**	X01-02	    Jim Duff	    26-Mar-2003
**	It was pointed out that I was ignoring the NMX bits in the 
**	file ID.  This was a Bad Thing (tm).  If you attempted to watch
**	a directory with non-zero NMX bits, the program would take out an
**	incorrect lock, and either watch the wrong file, or watch a
**	nonexistant "file".   The NMX bits have now been included
**	to form the correct file ID under all circumstances.
**
**	Additionally, the XQP always takes out the volume lock in
**	concurrent read mode and the serialization lock in protected write
**	mode.  Let's do the same thing.
**
**	Add some unsigned modifiers to variable definitions to match
**	the way VMS conventionally uses these variables.
**
**	X01-03	    Jim Duff	    15-Oct-2004
**	Correct a timing window that could hang up the code forever.
**
**	X01-04	    Jim Duff	    22-Oct-2007
**	Do some code standardization.
**
**	X01-05	    Jim Duff	    16-Jun-2009
**	Check that the parameter passed to us really is a directory.  Handle
**	directory specifications passed as a file (e.g. my_disk:[wrk]test.dir)
**	or directory specs passed as a dir (e.g., my_disk:[wrk.test]).  Add
**	some comments describing how to run the code.  Switch to New Starlet.
**
**      X01-06      Jim Duff        24-May-2013
**      Jim: This change is really by Hein van den Heuvel, who was
**	investigating using this stuff in a product.  He did a little more
**	testing than I did and turned up a failure scenario.  In addition he
**	taught me that the value block of the directory serialization lock
**	contains a sequence number that shows if the directory access actually
**	changed the directory.  I was unaware of this feature.  The day you
**	stop learning is the day you die :-)
**      Thanks Hein!
**
**      Hein: Check for Directory Changes, not just access.
**	Add rundown handler to attempt to get rid of the kernel locks which 
**	survive image rundown. 
** 
**	Jim: Also, handle control-y to ensure the exit handler is run if the 
**	code is being run interactively.
**
**--
*/

#include <stdio.h>
#include <stdlib.h>
#include <ssdef.h>
#include <stsdef.h>
#include <string.h>
#include <assert.h>
#include <lckdef.h>
#include <descrip.h>
#include <dvidef.h>
#include <jpidef.h>
#include <rms.h>
#include <iosbdef.h>
#include <iledef.h>
#include <efndef.h>
#include <iodef.h>
#include <stat.h>
#include <libclidef.h>
#include <lib$routines.h>
#include <starlet.h>

#define errchk_sig(arg) if (!$VMS_STATUS_SUCCESS(arg)) { \
			    (void)lib$signal(arg); \
			}
#define errchk_stop(arg) if (!$VMS_STATUS_SUCCESS(arg)) { \
			    (void)lib$stop(arg); \
			}

/*
** Declare a global structure to hold vaiables that the kernel mode
** routines use.
*/
static struct {
    struct {
	unsigned short int status;
	short int unused;
	unsigned long int lock_id;
    } plock_sb;
    struct {
	unsigned short int status;
	short int unused;
	unsigned long int lock_id;
	unsigned int fc_hdrseq;
	unsigned int fc_dataseq;
	unsigned int reserved[2];
    } clock_sb;
    unsigned int sts;
    unsigned int efn;
    unsigned char logname[12];
    unsigned short int fid;
    unsigned char nmx;
    unsigned char rvn;
} kmp;

static struct {
    int forward_link;
    void *handler_addr;
    int arg_count;
    int *condition_value_ptr;
    int condition_value;
} exit_descriptor;


/******************************************************************************/
static void drop_lock (void) {
/*
** This routine is called as an AST in kernel mode.
*/

static unsigned long int r0_status;

    /*
    ** Drop the serialization lock down to NULL.
    */
    r0_status = sys$enqw (EFN$C_ENF,
			  LCK$K_NLMODE,
			  &kmp.clock_sb,
			  LCK$M_CONVERT,
			  0,
			  0,
			  0,
			  0,
			  0,
			  0,
			  0);
    if (!$VMS_STATUS_SUCCESS (r0_status)) {
	kmp.sts = r0_status;
    }

    /*
    ** Set the event flag the main routine is waiting on.
    */
    r0_status = sys$setef (kmp.efn);
    if (!$VMS_STATUS_SUCCESS (r0_status)) {
	kmp.sts = r0_status;
    }
}


/******************************************************************************/
static int del_lock (void) {
/*
** This routine is called as an AST in kernel mode.
** Delete both locks.
*/

static unsigned long int r0_status;

    r0_status = SS$_NORMAL;
    if (kmp.clock_sb.lock_id != 0) {
	r0_status = sys$deq (kmp.clock_sb.lock_id,
			     0,
			     0,
			     0);
        if ($VMS_STATUS_SUCCESS (r0_status)) {
            kmp.clock_sb.lock_id = 0;
            if (kmp.plock_sb.lock_id != 0) {
		r0_status = sys$deq (kmp.plock_sb.lock_id,
				     0,
				     0,
				     0);
		if ($VMS_STATUS_SUCCESS (r0_status)) {
                    kmp.plock_sb.lock_id = 0;
		}
	    }
	}
    }
    return r0_status;
}


/******************************************************************************/
static void oob_handler (unsigned int *mask) {
/*
** Executed if the user types Ctrl-Y to ensure our exit handler fires.
*/
    (void)printf ("Aborted by Control-Y - no restart possible.\n");
    lib$enable_ctrl (mask);
    exit (EXIT_SUCCESS);
}


/******************************************************************************/
static void exit_handler (int *condition_value_ptr) {
/*
** Execute on image rundown.  Delete any kernel mode locks we have.
*/

static unsigned long int r0_status;

    r0_status = sys$cmkrnl (&del_lock,
			    0);
    if (r0_status != SS$_NOPRIV) {
        errchk_sig (r0_status);
    }
}


/******************************************************************************/
static int get_lock (void) {
/*
** This routine is executed in kernel mode
** Just get NULL locks for now.  Arm later.
*/

static unsigned long int r0_status;

static struct {
    unsigned char prefix[6];
    unsigned char logname[12];
} parent_resource;

static struct {
    unsigned char prefix[6];
    unsigned short int fid;
    unsigned char nmx;
    unsigned char rvn;
} child_resource;

static struct dsc$descriptor_s parent_resource_d =
				{ sizeof (parent_resource),
				  DSC$K_DTYPE_T,
				  DSC$K_CLASS_S,
				  (char *)&parent_resource };

static struct dsc$descriptor_s child_resource_d =
				{ sizeof (child_resource),
				  DSC$K_DTYPE_T,
				  DSC$K_CLASS_S,
				  (char *)&child_resource };


    /*
    ** Construct the parent and child resources.
    */
    (void)memcpy (parent_resource.prefix, "F11B$v", 6);
    (void)memcpy (parent_resource.logname, kmp.logname, 12);

    (void)memcpy (child_resource.prefix, "F11B$s", 6);
    child_resource.fid = kmp.fid;
    child_resource.nmx = kmp.nmx;
    child_resource.rvn = kmp.rvn;

    /*
    ** Obtain a NULL mode lock on the volume resource.
    */
    r0_status = sys$enqw (EFN$C_ENF,
			  LCK$K_NLMODE,
			  &kmp.plock_sb,
			  LCK$M_SYSTEM,
			  &parent_resource_d,
			  0,
			  0,
			  0,
			  0,
			  0,
			  0);
    if (!$VMS_STATUS_SUCCESS (r0_status)) {
	return r0_status;
    }

    /*
    ** Obtain the serialization lock for the file in NULL mode. 
    */
    r0_status = sys$enqw (EFN$C_ENF,
			  LCK$K_NLMODE,
			  &kmp.clock_sb,
			  LCK$M_SYSTEM | LCK$M_VALBLK,
			  &child_resource_d,
			  kmp.plock_sb.lock_id,
			  0,
			  0,
			  0,
			  0,
			  0);
    return r0_status;
}
			  

/******************************************************************************/
static int arm_lock (void) {
/*
** This routine is executed in kernel mode
** Have NULL locks.  Arm now.
*/

static unsigned long int r0_status;

    /*
    ** Obtain the serialization lock for the file in protected write mode. 
    ** Note we specify the blocking ast parameter as our drop_lock routine.
    */
    r0_status = sys$enqw (EFN$C_ENF,
			  LCK$K_PWMODE,
			  &kmp.clock_sb,
			  LCK$M_CONVERT | LCK$M_VALBLK,
			  0,
			  0,
			  0,
			  0,
			  &drop_lock,
			  0,
			  0);
    return r0_status;
}
			  

/******************************************************************************/
static int validate_dir (char *file) {
/*
** Use C's stat() function to make sure this filename is really a directory.
*/

static struct stat data;

    if (stat (file, &data) == 0) {
	if (S_ISDIR (data.st_mode)) {
	    return TRUE;
        }
    }
    return FALSE;
}


/******************************************************************************/
int main (void) {
/*
** Main routine.  Obtain the name of the directory the user wants to watch.
** Find the volume lock id of the device the file is on, and the FID of the
** file.  These pieces of info are required to format the resource names for
** the locks.
** Get to kernel to obtain the locks and then wait for the AST to fire to
** tell us to go ahead.
*/

static struct FAB fab;
static struct NAM nam;

static IOSB iosb;

static char *p;

static int r0_status;
static int i;
static int use_did = FALSE;
static int mode;
static int fc_hdrseq;
static int fc_dataseq;
static unsigned int mask = LIB$M_CLI_CTRLY;
static unsigned int old_mask;

static unsigned short int channel;

static unsigned char devlockname[64];

static ILE3 dviitms[] = { 64, DVI$_DEVLOCKNAM, devlockname, NULL,
			   0, 0, NULL, NULL };

static char file[NAM$C_MAXRSS];
static char es[NAM$C_MAXRSS];
static char rs[NAM$C_MAXRSS];

static struct dsc$descriptor_s file_d = { sizeof (file),
					  DSC$K_DTYPE_T,
					  DSC$K_CLASS_S,
					  file };

static const $DESCRIPTOR (prompt_d, "Directory to watch: ");
static $DESCRIPTOR (terminal_d, "TT:");



    /*
    ** Get the name of the directory the user wants to watch
    */
    r0_status = lib$get_foreign (&file_d, &prompt_d, &file_d);
    if (r0_status == RMS$_EOF) {
	exit (EXIT_SUCCESS);
    } else {
	errchk_stop (r0_status);
    }
    file[file_d.dsc$w_length] = '\0';

    /*
    ** Well, the directory better be of some substantial length...
    */
    if (file_d.dsc$w_length == 0) {
	(void)fprintf (stderr,
		       "ERROR: You must specify a directory to watch\n");
	exit (EXIT_FAILURE);
    }
    
    /*
    ** Set up a RMS file access block specifying the file name.  Attach a
    ** Name attributes block.
    */
    fab = cc$rms_fab;
    fab.fab$l_nam = &nam;
    fab.fab$l_fop = FAB$M_NAM;
    fab.fab$l_fna = file;
    fab.fab$b_fns = file_d.dsc$w_length;

    /*
    ** Set up a name attributes block
    */
    nam = cc$rms_nam;
    nam.nam$l_esa = es;
    nam.nam$b_ess = sizeof(es);
    nam.nam$l_rsa = rs;
    nam.nam$b_rss = sizeof(rs);

    /*
    ** Use RMS to parse the file to pick up any fields in the filename
    ** the user failed to specify.
    */
    r0_status = sys$parse (&fab);
    if (r0_status == RMS$_DNF) {
	(void)fprintf (stderr,
		       "ERROR: directory %-.*s not found\n",
		       file_d.dsc$w_length,
		       file_d.dsc$a_pointer);
	exit (EXIT_FAILURE);
    } else {
        errchk_stop (r0_status);
    }

    if ((nam.nam$l_fnb & NAM$M_NODE) != 0) {
        (void)fprintf (stderr,
		       "ERROR: This utility does not support node specs.\n");
	exit (EXIT_FAILURE);
    }

    /*
    ** If the name and type are specified, we will use the FID of this file
    ** if it exists (and if it's a directory).
    */
    if (((nam.nam$l_fnb & NAM$M_EXP_NAME) != 0) &&
        ((nam.nam$l_fnb & NAM$M_EXP_TYPE) != 0)) {
	/*
	** Use RMS to search for the file.  This fills in the NAM block, 
	** including the file id.
	*/
        r0_status = sys$search (&fab);
	if (r0_status == RMS$_FNF) {
	    (void)fprintf (stderr,
			   "ERROR: File %-.*s not found\n",
			   file_d.dsc$w_length,
			   file_d.dsc$a_pointer);
	    exit (EXIT_FAILURE);
	} else {
	    errchk_stop (r0_status);
	}
	if ((nam.nam$b_type != 4) ||
	    (strncasecmp (".DIR", nam.nam$l_type, 4) != 0)) {
	    (void)fprintf (stderr,
			   "ERROR: File must be .DIR type\n");
	    exit (EXIT_FAILURE);
	}
	use_did = FALSE;
    } else {
	use_did = TRUE;
    }

    switch (nam.nam$b_ver) {
	case 1:
	    /*
	    ** User specified ";" or ".".  Nothing to check...
	    */
	    break;
	case 2:
	    /*
	    ** Version length of 2.  The second character better be "1"...
	    */
	    p = (char *)nam.nam$l_ver + 1;
	    if (*p != '1') {
		(void)fprintf (stderr,
			       "ERROR: Version number of directoriess "
			       "must be ;1\n");
		exit (EXIT_FAILURE);
	    }
	    break;
	default:
	    /*
	    ** Version length > 2.  Let's just blow up.  Things like ";-0"
	    ** are rare.
	    */
	    (void)fprintf (stderr,
			   "ERROR: Version number of directoriess "
			   "must be ;1\n");
	    exit (EXIT_FAILURE);
	    break;
    }	

    /*
    ** Store the directory's file id in the global structure.
    */
    if (use_did) {
	kmp.fid = nam.nam$w_did_num;
	kmp.nmx = nam.nam$b_did_nmx;
	kmp.rvn = nam.nam$b_did_rvn;
    } else {
        kmp.fid = nam.nam$w_fid_num;
        kmp.nmx = nam.nam$b_fid_nmx;
        kmp.rvn = nam.nam$b_fid_rvn;
    }

    /*
    ** We *must* watch a directory.  Obviously, there are serialization locks
    ** for normal files too.  But we certainly don't want to end up watching
    ** a normal file.  So let's check for real if the file we are looking at
    ** is a directory.
    */
    if (!use_did) {
	/*
	** We only need to do this if we are using the FID as opposed to the
	** DID.  If we use the DID that RMS handed us, we already know it's a
	** directory.
	*/
	if (!validate_dir (file)) {
	    (void)fprintf (stderr,
			   "ERROR: XQP indicates %-.*s is not a directory\n",
			   file_d.dsc$w_length,
			   file_d.dsc$a_pointer);
	    exit (EXIT_FAILURE);
	}
    }

    /*
    ** Reuse our filename string descriptor
    */
    assert (sizeof (file) >= nam.nam$b_dev);
    (void)memcpy (file, nam.nam$l_dev, nam.nam$b_dev);
    file_d.dsc$w_length = nam.nam$b_dev;

    /*
    ** Call SYS$GETDVI to obtain the volume lock name.  This is used
    ** to construct the resource for the volume lock, which is the parent
    ** of all serialization locks.
    */
    r0_status = sys$getdviw (EFN$C_ENF,
			     0,
			     &file_d,
			     dviitms,
			     &iosb,
			     0,
			     0,
			     0);
    errchk_stop (r0_status);
    errchk_stop (iosb.iosb$w_status);

    /*
    ** SYS$GETDVI returns a 64 byte string.  The first byte indicates if
    ** the volume is mounted system, group, or private.  We can ignore it.
    ** The next 12 bytes are what we are interested in.
    */
    for (i = 1; i <= 12; i++) {
	kmp.logname[i-1] = devlockname[i];
    }

    /*
    ** Get an event flag to wait on, and clear it
    */
    r0_status = lib$get_ef (&kmp.efn);
    errchk_stop (r0_status);

    /*
    ** We are about to acquire locks in kernel mode.  Set up an exit handler
    ** to delete them on image exit.  Additionally, set up a OOB handler for
    ** control-c and control-y to make sure experimenters running this
    ** interactively can't shoot themselves in the foot.
    */
    exit_descriptor.handler_addr = &exit_handler;
    exit_descriptor.arg_count = 1;
    exit_descriptor.condition_value_ptr = &exit_descriptor.condition_value; 

    r0_status = sys$dclexh (&exit_descriptor);
    errchk_stop (r0_status);

    r0_status = lib$getjpi (&JPI$_MODE,
			    0,
			    0,
			    &mode,
			    0,
			    0);
    errchk_sig (r0_status);

    if (mode == JPI$K_INTERACTIVE) {
	r0_status = lib$disable_ctrl (&mask,
				      &old_mask);
	errchk_sig (r0_status);

        r0_status = sys$assign (&terminal_d,
				&channel,
				0,
				0,
				0);
	errchk_sig (r0_status);

	r0_status = sys$qiow (EFN$C_ENF,
			      channel,
			      IO$_SETMODE|IO$M_CTRLYAST,
			      &iosb,
			      0,
			      0,
			      &oob_handler,
			      (__int64)&old_mask,
			      0,
			      0,
			      0,
			      0);
	errchk_sig (r0_status);
	errchk_sig (iosb.iosb$w_status);
    }

    /*
    ** We also need to get rid of the lock if the user hits control-y.
    */
    /*
    ** Get to kernel and lock the resources in NULL mode and save initial
    ** cache sequence values
    */
    r0_status = sys$cmkrnl (&get_lock,
			    0);
    if (r0_status == SS$_NOPRIV) {
	(void)fprintf (stderr,
		       "ERROR: Requires CMKRNL privilege to run\n");
	/*
	** We can cancel our exit handler as we didn't get the locks.
	*/
	r0_status = sys$canexh (&exit_descriptor);
	errchk_stop (r0_status);
	exit (EXIT_FAILURE);
    } else {
	errchk_sig (r0_status);
    }

    /*
    ** Save sequence values.
    */
    fc_hdrseq  = kmp.clock_sb.fc_hdrseq;
    fc_dataseq = kmp.clock_sb.fc_dataseq;

    while (TRUE) { 
	/*
	** Clear the event flag we are waiting on.
	*/
        r0_status = sys$clref (kmp.efn);
        errchk_stop (r0_status);

	/*
	** Switch to kernel and convert locks and arm blocking ast.
	*/
	r0_status = sys$cmkrnl (&arm_lock, 0);
        errchk_stop (r0_status);

        /*
        ** Need to check value block AFTER getting the lock back
        ** $CREATE will blast us before actual create.
	** The altered dataseq is returned in the lock value block on release.
        ** ( see: VMS File Systems Internals' by Kirby McCoy, page 324)
        */
        if (fc_dataseq != kmp.clock_sb.fc_dataseq) {
	    break;
	}

        /*
        ** Wait on the event flag.  The blocking AST which drops the locks will
        ** set this flag so the program exits IF the directory is changed.
        */
        r0_status = sys$waitfr (kmp.efn);
        errchk_sig (r0_status);
    }

    /*
    ** Delete the locks.
    */
    r0_status = sys$cmkrnl (&del_lock, 0);
    errchk_stop (r0_status);

    /*
    ** We just did what our exit handler would have done, so no need for it
    ** now.
    */
    r0_status = sys$canexh (&exit_descriptor);
    errchk_stop (r0_status);

    if (mode == JPI$K_INTERACTIVE) {
	r0_status = sys$dassgn (channel);
	errchk_sig (r0_status);

	r0_status = lib$enable_ctrl (&old_mask,
				     0);
	errchk_sig (r0_status);
    }

    /*
    ** For good form's sake, free the event flag.
    */
    r0_status = lib$free_ef (&kmp.efn);
    errchk_sig (r0_status);
}
